import warnings 
from collections import defaultdict
from functools import partial

import numpy as onp
try: 
    import jax
    import jax.numpy as np
except: 
    np=onp
    warnings.warn('Jax import failed.')



def diff_solve(f, gradf, hessf=None, ff=None, nr=None,verbose=False, 
            step=None, soptions={}, iterate=None, itoptions={},
            solve=None, svoptions={},init=None, inoptions={}):

    ##########################################################################
    ## initialize strategy
    ##########################################################################

    if nr is not None: nr = -nr

    ## State Handler
    if callable(init):
        pass 
    else:
        inoptions.setdefault('verbose',verbose)
        inoptions.setdefault('nr',nr)
        init = _init_init_basic(f,gradf,**inoptions)

    ## Iterator 
    if callable(iterate): 
        pass
    else:
        itoptions.setdefault('nr',nr)
        itoptions.setdefault('verbose',verbose)
        iterate = iterate_init(f, gradf, hessf, **itoptions)

    ## Stepper
    if callable(step): pass
    else:
        if not soptions: # soptions is empty
            pass
        step = lambda state0, **opts:  [state0]

    ## Solver
    if callable(solve): pass
    else:
        if not soptions: # soptions is empty
            pass
        solve = init_solve_basic(init,step,iterate,False)

    ##########################################################################
    ## Solver
    ##########################################################################
    # def solve(x0, params={}, ff=None):
    #     a = params
    #     x, a, state0 = init_solve(x0, a, ff)

    #     for step,state in enumerate(stepper(state0)):

    #         if verbose: print( 'step: ', step + 1)
            
    #         x, _ , _ = iterate(x, a, state)

    #     return x
    # svoptions = {}
    # _solve = solve(init, step, iterate,**svoptions)

    return solve


def path_solve(f, gradf, ff=None, nr=None, verbose=False, nosteps=1,
                soptions={}, itoptions={},svoptions={},inoptions={}):

    if nr is not None: nr = -nr

    ## State Handler
    init = _init_init_path(f,gradf,verbose=verbose,**itoptions)

    ## Iterator 
    iterate = _init_iter_path(f, gradf,verbose=verbose, **itoptions)

    ## Stepper
    step = lambda state0, **opts:  [state0]
    
    step = _init_incr_path(nosteps,verbose=verbose)


    ## Solver
    solve = _init_solv_path(init,step,iterate,verbose=verbose)

    return solve



def init_solve_basic(init,incr,iterate,verbose=False):

    def solve(x0, params={}, ff=None):
        a = params
        x, a, state0 = init(x0, a, ff)

        for i,state in enumerate(incr(state0)):

            if verbose: print( 'step: ', i + 1)
            
            x, _ , _ = iterate(x, a, state)

        return x

    return solve


def _init_init_basic(f, gradf, verbose=False,nr=None,**kwargs):

    def init(x0, a={}, ff=None):
        if verbose: print('Initializing iterative solve.')
        state = {}
        x = x0

        f0 = f(x0, **a)[:nr]
        state['fi'] = f0
        if verbose: print('\nf0: ',np.around(f0,3))

        if ff is None: ff = np.zeros(f0.shape)
        elif callable(ff): ff = ff(x0)[:nr]
        if verbose: print('\nff: ',np.around(ff,3))

        state.setdefault('df',  ff-f0)
        state.setdefault('ff',     ff)
        state.setdefault('nr',     nr)
        return x, a, state
    
    return init

def iterate_init(f, gradf, hessf=None, nr=None, maxiter=20,
                 tol=1e-3,loss=None,verbose=False,jit=True,**kwargs):
    
    if loss is None:
        loss = np.linalg.norm
    
    def update(x, a, state):
        if verbose: print('x: ', x)
        df = state['df']
        if verbose: print('df: ',df.T)

        gradfi = gradf(x,**a)[:nr,:nr]
        if verbose: print('gradf: \n',gradfi)

        dx = np.linalg.solve(gradfi, df)[:,0]

        if nr is not None:
            dx = np.pad(dx, (0,-nr), 'constant') 
        if verbose: print('dx: ',dx.T)
        x = x + dx

        state['fi'] = f(x, **a)[:nr]
        # state['df'] = df
        return x, state

    if jit:
        update = jax.jit(update)
        print('Iteration updater jit compilation successful.')
    
    def iterate(x0, a, state):
        x = x0
        for itr in range(maxiter):
            if verbose: print('iteration: ',itr)
            state['df'] = state['ff'] - state['fi']
            
            if loss(state['df']) <= tol: 
                return x, None, state

            x,  state = update(x, a, state)

        if verbose:
            msg = ("Failed to converge after %d iterations, value is %s."
                % (itr + 1, loss(state['df'])))
            raise RuntimeError(msg)
        
    return iterate



def _init_init_path(f, gradf, verbose=False,nr=None,**kwargs):

    def init_solve(x0, a=(), ff=None):
        if verbose: print('Initializing iterative solve.')
        state = {}
        x = x0

        f0, a = f(x0, *a)
        f0 = f0[:nr]
        state['fi'] = f0
        if verbose: print('\nf0: ',np.around(f0,3))

        if ff is None: ff = np.zeros(f0.shape)
        elif callable(ff): ff = ff(x0)[:nr]
        if verbose: print('\nff: ',np.around(ff,3))

        state.setdefault('df',  ff-f0)
        state.setdefault('ff',     ff)
        state.setdefault('nr',     nr)

        return x, a, state
    
    return init_solve

def _init_iter_path(f, gradf, hessf=None, nr=None, maxiter=20,
                    tol=1e-3,loss=None,verbose=False,jit=True,**kwargs):
    
    if loss is None:
        loss = np.linalg.norm
    
    def update(x, a, state):
        if verbose: print('x: ', x)
        df = state['df']
        if verbose: print('df: ',df.T)

        gradfi, _ = gradf(x,*a)
        gradfi = gradfi[:nr,:nr]
        if verbose: print('gradf: \n',gradfi)

        dx = np.linalg.solve(gradfi, df)

        if nr is not None:
            dx = np.pad(dx, (0,-nr), 'constant')

        if verbose: print('dx: ',dx.T)
        x = x + dx

        fi, a = f(x, *a)

        state['fi'] = fi[:nr]
        return x, a, state

    # if jit:
    #     try:
    #         update = jax.jit(update)
    #     except:
    #         pass
    
    def iterate(x0, a, state):
        x = x0
        for itr in range(maxiter):
            if verbose: print('iteration: ',itr)
            state['df'] = state['ff'] - state['fi']

            if loss(state['df']) <= tol: 
                return x, a, state

            x, a, state = update(x, a, state)

        msg = ("Failed to converge after %d iterations, value is %s."
            % (itr + 1, loss(state['df'])))
        # raise RuntimeError(msg)
        warnings.warn(msg)
        return x, a, state
        

    return iterate

def _init_incr_path(_,**kwds):

    def step(state0,nosteps):
        ff = state0['ff']
        df = state0['df']
        state = state0.copy()

        for i in range(1,abs(int(nosteps))+1):
            # state.update(ff=i*ff/nosteps)
            # state.update(df=i*df/nosteps)
            yield state


    return step

def _init_solv_path(init,incr,iterate,verbose=False):

    def solve(x0, params={}, ff=None,nostep=1):
        a = params
        x, a, state0 = init(x0, a, ff)

        for i,state in enumerate(incr(state0,nostep)):

            if verbose: 
                print( 'step: ', i + 1)
                # print( 'ff : ', state['ff'])
                # print( 'df : ', state['df'])
            # print(state['ff'], x)
            x, a , _ = iterate(x, a, state)
            # print(a)

        return x, a

    return solve




from functools import partial

import jax
import jax.numpy as jnp

def Truss001(jit=False,**kwds):
    def f(u, x0, E, A, q0=0.0, params={} ):
        DX = x0[1,0] - x0[0,0]
        DY = x0[1,1] - x0[0,1]
        L = jnp.linalg.norm([DX,DY])

        v = (u[2]-u[0])*DX/L + (u[3]-u[1])*DY/L + ((u[2]-u[0])**2 + (u[3]-u[1])**2)/(2*L)
        dv = jnp.array([[-DX/L - (u[2]-u[0])/L],
                        [-DY/L - (u[3]-u[1])/L], 
                        [ DX/L + (u[2]-u[0])/L],
                        [ DY/L + (u[3]-u[1])/L]])
        dv = dU@dv
        e = v/L 
        q = E*A*e + q0
        return (dv*q, None)

    def Df(u, x0, E, A, q0=0.0):
        pass

    Df = jax.jit(partial(Df,**kwds))
    f = partial(f,**kwds)
    if jit: f = jax.jit(f)
    params = {}

    Space = object()
    Space.f = f 
    Space.Df = Df 
    Space.maps = ( params, [4] )
    
    return Space



import collections
import jax
from jax import lax
import jax.numpy as np
from scipy.optimize import fsolve
from functools import partial
# from .roots import path_solve

def return_map(E, H0, C, e0=0.0, ep0=0.0, sigma0=0.0, q0=0.0, Y0=0.0, AD=False, H=None,init=False,linearize=False, **kwds):
    if H is None: H = lambda Dep: H0

    def g(Dep, sigma_trial, q_i, Y_i):
        return abs(sigma_trial - q_i) - (E + C)*Dep - Y_i - H( sigma_trial - (E + C)*Dep )*Dep
        
    if linearize:
        def solve_g(sigma_trial, q_i, Y_i):
            return (abs(sigma_trial-q_i) - Y_i )/(E+C+H0)
    else:
        def solve_g(sigma_trial, q_i, Y_i):
            return fsolve(g, 0., args=(sigma_trial, q_i, Y_i))[0]
    
    def step(De, e_i, ep_i, sigma_i, q_i, Y_i,*args):  # De, e_i, sigma_i, q_i, Y_i, ep_i 
        e_n = e_i + De                    # New total strain at the end of this step
        sigma_trial = sigma_i + E * De    # Trial stress
        xi_trial = sigma_trial  - q_i
        f_trial = abs(xi_trial) - Y_i

        if f_trial < 0.0: # Elastic step
            sigma_n = sigma_trial
            Et = E
            Y_n  = Y_i
            ep_n = ep_i
            q_n  = q_i
            
        else:
            Dep = solve_g(sigma_trial, q_i, Y_i)
            ep_n = ep_i + Dep * np.sign(xi_trial)
            sigma_n = E*(e_n - ep_n)
            Y_n =  Y_i + H(Y_i) * Dep
            Et = E*(C + H(Y_i))/( E + C + H(Y_i) )
            q_n = q_i + C*Dep*np.sign(xi_trial)

        return e_n, ep_n, sigma_n, q_n, Y_n, Et

    # step = jax.jit(step, static_argnums=(0,2,1,3,4,5))

    if init:
        State = collections.namedtuple('State', 'e, ep, sigma, q, Y, Et', defaults=(e0, ep0, sigma0, q0, Y0, E))
        return step, State
    return step




"""
=====================================================

.. currentmodule:: anabel

"""
# from .version import __version__
__version__ = '0.0.0'



try:
    from ._anabel import execute_model  # noqa
except ImportError:
    def execute_model(args):
        return max(args, key=len)

try:
    import jaxlib
except ImportError:
    COMPILER = None
else:
    COMPILER = jaxlib



# tensorlib = tensor.numpy_backend()
# default_backend = jax.numpy


# def get_backend():
#     """
#     Get the current backend and the associated optimizer
#     Example:
#         >>> import pyhf
#         >>> pyhf.get_backend()
#         (<pyhf.tensor.numpy_backend.numpy_backend object at 0x...>, <pyhf.optimize.opt_scipy.scipy_optimizer object at 0x...>)
#     Returns:
#         backend, optimizer
#     """
#     global tensorlib
#     global optimizer
#     return tensorlib


# @events.register('change_backend')
# def set_backend(tensor_library, compiler='jaxlib'):
#     """
#     Set the backend and the associated optimizer
#     Args:
#         tensor_library (`str`): One of the supported pyhf backends: numpy, jax
#         compiler       (`str`): Optional compiler 
#     Returns:
#         None
#     """
#     global tensor_library
#     global compiler

#     if isinstance(tensor_library, (str, bytes)):
#         if isinstance(tensor_library, bytes):
#             tensor_library = tensor_library.decode("utf-8")
#         tensor_library = tensor_library.lower()

#     # need to determine if the tensor_library changed 
#     tensorlib_changed = bool(
#         (tensor_library.name != tensorlib.name) | (tensor_library.precision != tensorlib.precision)
#     )
#     # set new backend
#     tensorlib = tensor_library
#     # trigger events
#     if tensorlib_changed:
#         events.trigger("tensorlib_changed")()
#     # set up any other globals for backend
#     tensorlib._setup()


from .core import *
from anabel import models
import io
# from anabel import numeric
# from emx.graphics import *
# from .fem import *
# import emx.analysis
# import emx.utilities
# from emx.tests import *
# from emx.tensorelements import *




from functools import partial
from fractions import Fraction

import jax.numpy as jnp



def lagrange(i=None,n=None,p=None,x=None):
    r"""
    Lagrange polynomial interpolation.
    Given two 1-D arrays `x` and `w,` returns the Lagrange interpolating
    polynomial through the points ``(x, w)``.
    Warning: This implementation is numerically unstable. Do not expect to
    be able to use more than about 20 points even if they are chosen optimally.
    Parameters
    ----------
    x : array_like
        `x` represents the x-coordinates of a set of datapoints.
    w : array_like
        `w` represents the y-coordinates of a set of datapoints, i.e. f(`x`).
    Returns
    -------
    lagrange : `numpy.poly1d` instance
        The Lagrange interpolating polynomial.
    Examples
    --------
    Interpolate :math:`f(x) = x^3` by 3 points.
    >>> from scipy.interpolate import lagrange
    >>> x = np.array([0, 1, 2])
    >>> y = x**3
    >>> poly = lagrange(x, y)
    Since there are only 3 points, Lagrange polynomial has degree 2. Explicitly,
    it is given by
    .. math::
        \begin{aligned}
            L(x) &= 1\times \frac{x (x - 2)}{-1} + 8\times \frac{x (x-1)}{2} \\
                 &= x (-2 + 3x)
        \end{aligned}
    >>> from numpy.polynomial.polynomial import Polynomial
    >>> Polynomial(poly).coef
    array([ 3., -2.,  0.])

    """

    if x is not None: 
        M = len(x)
    else:
        M = n 
        x = jnp.linspace(-1,1,n)
    cf = jnp.ones(1)
    for k in range(M):
        if k == i: continue
        fac = x[i]-x[k]
        cf = jnp.convolve(cf,   jnp.array([1.0, -x[k]])/fac)
    return cf



import os

from anabel import COMPILER 

compiler = COMPILER


try:
    from ruamel_yaml import YAML
except ImportError:
    yaml = None
else: 
    yaml = YAML(typ='safe')


def _yaml_load(filepath):
    with open(filepath) as f: model_dict = yaml.load(f)
    return model_dict

def load(filepath):
    """Load from a serialized data file"""
    _, file_ext = os.path.splitext('/path/to/somefile.ext')
    
    if file_ext in ['.yml', '.yaml']: 
        return _yaml_load(filepath)

def dump(filepath): pass

def compile(func,inputs):
    if compiler is not None: 
        pass




"""
A collection of functions for composing graphs of finite elements.
"""

from collections import namedtuple, deque
from collections import MutableMapping 

# from anabel.models import _linear_wire, model


def compose(elements, tree, key=None, _built=False, node='ElemSpace'):
    """Apply function composition over a computational graph."""
    if key is None: 
        key = list(tree[node].keys())[0]
        tree = list(tree[node].values())[0]

    if node in tree and (not _built): # build grandchild elements
        tree[node] = {
            hkey: compose( elements, tree[node][hkey], key=hkey) 
            for hkey in [eli for eli in tree[node]] }

        return compose(elements, tree, key, True) # build child element

    else: return elements[key](**tree)

Create_Model = compose

def merge(d1, d2):
    """
    Merge two dicts of dicts recursively.
    If either mapping has leaves that are not instances 
    of a `MutableMapping`, the second's leaf overwrites 
    the first's.
    """
    for k, v in d1.items():
        if k in d2:
            if all(isinstance(e, MutableMapping) for e in (v, d2[k])):
                d2[k] = merge(v, d2[k])
    d3 = d1.copy()
    d3.update(d2)
    return d3

def dict_depth(d):
    queue = deque([(id(d), d, 1)])
    memo = set()
    while queue:
        id_, o, level = queue.popleft()
        if id_ in memo: continue
        memo.add(id_)
        if isinstance(o, dict):
            queue += ((id(v), v, level + 1) for v in o.values())
    return level

Element = namedtuple('Element', 'f, Df, params')

def state_wrapper(element):
    return Element(
        lambda **kwds: (element.f(**kwds), None),
        element.Df ,
        element.state
    )

# Model = {
#     'basic-linear': _linear_wire
# }

###############################################################
# Combinators
###############################################################
def Z(f): 
    """Z-combinator
    $$\lambda f .(\lambda x .f(\lambda v .((x x) v)))(\lambda x . f(\lambda v .((x x) v)))$$"""
    return (lambda x: f(lambda v: x(x)(v)))(lambda x: f(lambda v:x(x)(v)))


def step (f):
    def incr (originalValue, approx):
        improvedApprox = (approx + (originalValue / approx)) / 2;

        # How far off the mark we are.
        discrepancy = abs(originalValue -  improvedApprox * improvedApprox)

        # Termination condition
        if discrepancy < 0.00001:
            return improvedApprox
        return f(originalValue, improvedApprox)
    return incr
